"""
.. module:: pyFC
   :synopsis: Fractal Cube generator, analyzer, and visualizer. Lognormal fractal cubes are constructed by the iterative method of
Lewis & Austin (2002).

.. moduleauthor:: Alexander Y. Wagner <alexaner.y.wagner@gmail.com>

"""
from utils import *

